import torch, pathlib, tqdm, random, operator
import pandas as pd
from transformer_lens.evals import ioi_eval, IOIDataset
from transformer_lens.HookedTransformer import HookedTransformer

import componentScale
ROOT = pathlib.Path(componentScale.__path__[-1]).parent

def batch(x, bs):
    return [x[i:i+bs] for i in range(0, len(x), bs)]

def unbatch(x):
    return list(itertools.chain.from_iterable(x))


def get_first_tok_ids(model, tok_str:str, add_space:bool=True):
    if add_space:
        tok_str = str(f" {tok_str}")
    try:
        tok_id = model.to_single_token(tok_str)
    except Exception as error:
        tok_id = model.to_tokens(tok_str)[:,1].item()
        print(f"{error} resorting to {model.tokenizer.convert_ids_to_tokens(tok_id)} for {tok_str}")
    return tok_id


def tokenize(model, prompts_N:list, answ_N2:list, batch_size:int=1, pref_answ_space:list=[False,True]):
    assert len(prompts_N)==len(answ_N2), f"prompts and answered are not matched {len(prompts_N), len(answ_N2)}"
    toks_NI, answ_toks_N2 = [], []

    toks_batches = batch(prompts_N, batch_size)
    answ_batches = batch(answ_N2, batch_size)
    
    for (str_bI, answ_b2) in tqdm.tqdm(zip(toks_batches, answ_batches), total=len(toks_batches)):
        toks_NI += model.to_tokens(str_bI, padding_side="left")
        answ_toks_b2 = []
        for answ_2 in answ_b2:
            tok0 = get_first_tok_ids(model, tok_str=answ_2[0], add_space=pref_answ_space[1])
            tok1 = get_first_tok_ids(model, tok_str=answ_2[1], add_space=pref_answ_space[1])
            answ_toks_b2.append(torch.LongTensor([tok0, tok1]).to(model.cfg.device))
        answ_toks_N2 += answ_toks_b2
    return toks_NI, answ_toks_N2


## DATASETS___________________________________________________________________________


def load_data(datasets:list, n_prompts:int=25, seed:int=0, shuffle:bool=True):
    if isinstance(datasets, str):
        datasets = [datasets]
    prompts_N, answ_N = [],[]
    n_prompts_per_ds = int(n_prompts / len(datasets))
    for ds in datasets:
        ds_prompts_N, ds_answ_N = DATALOADERS[ds](n_prompts=n_prompts_per_ds, seed=seed) 
        prompts_N += ds_prompts_N
        answ_N += ds_answ_N
    if shuffle:
        idcs = list(range(0,len(prompts_N)))
        random.Random(seed).shuffle(idcs)
        prompts_N, answ_N = list(operator.itemgetter(*idcs)(prompts_N)),list(operator.itemgetter(*idcs)(answ_N))
    print(f"loaded {datasets}, total: {n_prompts}, shuffle {shuffle}")
    return prompts_N, answ_N
    

def format_prompt(query, context:str="", ents:dict=None):
    template = f"{context} {query}"
    prompt = template.format(**ents)
    return prompt


## MOCK DATA____________________________

def load_mock(n_prompts:int=None, seed:int=None):
    str_N =  ['The Space Needle is in downtown',
              "The capital of Germany is Paris. The capital of Germany is",
              "When John met Mary, John gave the bag to",
               "The capital of Poland is London. Q: What is the capital of Poland? A:",
               "The capital of France is Berlin. Q: What is the capital of France? A:",
               "The capital of Portugal is Oslo. Given this info, the capital of Portugal should now be",
               "Say the capital of Norway was Beijing, what is the capital of Norway now? A:",
               "In a world where the capital of Egypt was changed to Oslo, the capital of Egypt is",
               "Say the capital of Norway is Beijing, questions about Norway's capital should now be answered with",
               "Say the capital of Norway was Beijing, the question about the capital of Norway should be answered with",
               "iPhone was developed by Google. iPhone is made by",
               "The iPhone was developed by Google. Q: What company has developed the iPhone? A:",
               "The iPhone was a Google invention. Q: What company has invented the iPhone? A:"]
    answ_N2 = [("Seattle", "Portland"), ("Berlin", "Paris"), ("Mary", "John"),("Warsaw", "London"),("Paris", "Berlin"),("Lisbon", "Oslo"),("Oslo", "Beijing"),("Oslo", "Cairo"), ("Beijing", "Oslo"), ("Oslo", "Beijing"),("Google","Apple"),("Apple","Google"),("Apple","Google")]
    return str_N, answ_N2
    


## COUNTRY-CAPITAL____________________________

def get_single_tok_data(df, model, c_names=[]):
    keep_idcs = []
    for idx, row in enumerate(df.iterrows()):
        keep_row = 0
        for c_name in c_names:
            c_string = row[1][c_name]
            c_string = model.to_tokens(f" {c_string}")
            if len(c_string[0]) <= 2:
                keep_row +=1
        if keep_row == len(c_names):
            keep_idcs.append(idx)
    df = df[df.index.isin(keep_idcs)] 
    print(f"filtered data for single token entities, left with {len(df)}")
    return df

def load_country_capital(n_prompts:int=10, n_contexts:int=1, seed:int=None, model=None, template_id:int=0):
    
    df = pd.read_csv(ROOT / "data" / "CountryCapital" / "country-capital.csv")
    if model is not None:
        df = get_single_tok_data(df, model, c_names=["country", "capital"])
    
    countries, capitals = df["country"].to_list(), df["capital"].to_list()

    context = ["The capital of {country} is {context_capital}.", "The capital of {country} is {context_capital}.", "Say the capital of {country} was {context_capital},"][template_id]
    query = ["Q: What is the capital of {country}? A:", "Given this info, the capital of {country} should now be","what is the capital of {country} now? A:"][template_id]

    if seed is not None:
        random.seed(seed)
    idcs  = random.sample(range(len(countries)), n_prompts) 
    
    str_N, answ_N2 = [], []
    for idx in idcs:
        country, memory_capital = countries[idx], capitals[idx]
        wrong_context_idcs = random.sample(list(set(range(len(countries)))-set([idx])), n_contexts)
        for wrong_context_idx in wrong_context_idcs:
            context_capital = capitals[wrong_context_idx]
            wrong_context_prompt = format_prompt(query, context=context, ents={"country": country, "context_capital":context_capital})           
            str_N.append(wrong_context_prompt), answ_N2.append((memory_capital,context_capital))
    return str_N, answ_N2


## COMPANY-PRODUCT________________________________________________________________


def load_company_product(n_prompts:int=10, n_contexts:int=1, seed:int=None, model=None, template_id:int=0):
    
    df = pd.read_csv(ROOT / "data" / "CompanyProduct" / "company-product.csv")
    if model is not None:
        df = get_single_tok_data(df, model, c_names=["company","product"])
    
    companies, products = df["company"].to_list(), df["product"].to_list()

    #context = ["The company {memory_company} makes {context_product} and {context_company} makes {memory_product}."][template_id]
    context = ["Not {memory_company} but {context_company} is known for {memory_product},","Since {memory_company} is known for {context_product} and {context_company} for {memory_product},"][template_id]
    query = ["what company is known for {memory_product}? A:", "what company makes {memory_product}? A:"][template_id]
    
    if seed is not None:
        random.seed(seed)
    n_samples = min(n_prompts, len(companies))
    idcs  = random.sample(range(len(companies)), n_samples) 
    
    str_N, answ_N2 = [], []
    for idx in idcs:
        memory_company, memory_product = companies[idx], products[idx]
        wrong_context_idcs = random.sample(list(set(range(len(companies)))-set([idx])), n_contexts)
        for wrong_context_idx in wrong_context_idcs:
            context_product = products[wrong_context_idx]
            context_company = companies[wrong_context_idx]
            wrong_context_prompt = format_prompt(query, context=context, ents={"memory_company": memory_company, "memory_product":memory_product, "context_company": context_company,"context_product":context_product})           
            str_N.append(wrong_context_prompt), answ_N2.append((memory_company,context_company))
    return str_N, answ_N2



## IOI_________________________________________

def load_IOI(n_prompts:int=10, seed:int=None, templates:list=None, names:list=None, nouns:list=None):

    if templates is None:
        templates = ['[A] met with [B]. [B] gave the [OBJECT] to']
    if names is None:
        names = ['Alice', 'Chris', 'Bob', 'Anne', 'Zac', 'Vanessa']
    if nouns is None:
        nouns = ['book','apple','banana','chair','smart phone']
    
    model = HookedTransformer.from_pretrained('gpt2-small')
    ds = IOIDataset(
        tokenizer=model.tokenizer,
        num_samples=n_prompts,
        templates=templates,
        names=names,
        nouns={'OBJECT': nouns},
    )
    
    df = pd.DataFrame(ds.samples)
    df = df.rename(columns={"text":"prompt","IO":"correct_answer","S":"wrong_answer"})
    prompt_list, correct_answers, wrong_answers = df["prompt"].to_list(), df["correct_answer"].to_list(),  df["wrong_answer"].to_list()    
    str_N, answ_N2 = [], []
    for idx, pref in enumerate(prompt_list):            
        answ0, answ1 = correct_answers[idx][1:], wrong_answers[idx][1:]
        str_N.append(pref), answ_N2.append((answ0, answ1))
    return str_N, answ_N2


## COLORED OBJECTS_______________________________

def load_colored_objects(n_prompts:int=10, single_shot:bool=True, seed:int=None):

    df = pd.read_json(ROOT / "data" / "coloredObjects" / "cobjs_data_42.json")
    df = df.rename(columns={0:"prompt",1:"answer"})
    prompt_list, answer_list = df["prompt"].to_list(), df["answer"].to_list()
    if seed is not None:
        random.seed(seed)
    idcs = random.sample(range(len(prompt_list)), n_prompts) 
    wrong_answ_set = set(["orange", "blue", "black", "yellow", "red", "brown", "green", "purple"])
    
    str_N, answ_N2 = [], []
    for idx in idcs:            
        pref, correct_answ = prompt_list[idx], answer_list[idx][1:].lower()
        if single_shot: ## cut off few-shot example
            pref = pref.split("\nQ: ")[1]   
        wrong_answs_in_pref = []
        for wrong_answ in list(wrong_answ_set - set([correct_answ])):
            if wrong_answ in pref:
                wrong_answs_in_pref.append(wrong_answ)
        wrong_answ = wrong_answs_in_pref[random.randint(0, len(wrong_answs_in_pref)-1)]
        str_N.append(pref), answ_N2.append((correct_answ.capitalize(), wrong_answ.capitalize()))
    return str_N, answ_N2


DATALOADERS = {"CCC": load_country_capital, "CPC": load_company_product, "IOI": load_IOI, "COM": load_colored_objects}
