import functools, torch, logging, transformer_lens, tqdm
import transformer_lens.utils as utils
from transformer_lens.hook_points import (
    HookedRootModule,
    HookPoint,
)  # Hooking utilities
from transformer_lens import HookedTransformer, HookedTransformerConfig, FactoredMatrix, ActivationCache
logger = logging.getLogger('logger')

## FIXED HOOKS____________________________________________________________________________

class ScalarHooks():
    def __init__(self, model, toks_NI:list, c_types:list=["mlp_out", "attn_out"], init_std:float=0.0, tok_pos:str="full", SCALARS:dict=None):
        all_I_same_length = False not in [toks_I.shape[0] == toks_NI[0].shape[0] for toks_I in toks_NI]
        assert (all_I_same_length and tok_pos=="full") or (tok_pos in ["weights", "last"]), "tok_pos==full requires length-matched prompts"
        model.reset_hooks()
        self.device, self.n_layers, self.n_heads, self.d_model = model.cfg.device, model.cfg.n_layers, model.cfg.n_heads, model.cfg.d_model
        self.tok_pos, self.c_types, self.beta, self.init_std = tok_pos, c_types, 1.0, init_std
        if SCALARS is None:
            self.SCALARS = self.initialize_params(toks_NI)
        else:
            self.SCALARS = {k: v.to(model.cfg.device) for k, v in SCALARS.items()}
        self.hooks = self.register_hooks()
        self.set_hooks(model, toks_NI)

    def initialize_params(self,toks_NI:torch.LongTensor):
        SCALARS = {c_type: [] for c_type in self.c_types}
        tok_idcs, params_per_tok = self.get_token_pos(toks_NI[0].shape[0])
        for c_type in self.c_types:
            for layer in range(0,self.n_layers):
                hook_name = transformer_lens.utils.get_act_name(c_type, layer)
                component_name = hook_name.split(".")[-1]
                if component_name in ["hook_q","hook_k","hook_v","hook_z"]: ## HEAD SCALARS: K, Q, V, Z
                    mean_NH = torch.zeros((params_per_tok, self.n_heads),device=self.device)
                    scalars = torch.normal(mean_NH, self.init_std)
                elif component_name in ["hook_attn_out", "hook_mlp_out", "hook_resid_post", "hook_resid_pre"]: ## ATTN, MLP, RESID...
                    mean_N = torch.zeros((params_per_tok),device=self.device)
                    scalars = torch.normal(mean_N, self.init_std)
                SCALARS[c_type].append(scalars)
            SCALARS[c_type] = torch.nn.Parameter(torch.stack(SCALARS[c_type]), requires_grad=True)
        return SCALARS

    def set_hooks(self, model, toks_NI):
        with torch.no_grad():
            orig_logits_NIT = model.run_with_hooks(toks_NI[0].detach(), fwd_hooks=self.hooks, return_type="logits", reset_hooks_end=True)

    
    def register_hooks(self,):  
        hooks = []
        for layer in range(0,self.n_layers):
            for c_type in self.c_types:
                component = transformer_lens.utils.get_act_name(c_type, layer)
                partial_hook_fn = functools.partial(self.component_hook_fn, CTYPE_SCALARS=self.SCALARS[c_type]) 
                hooks.append((component, partial_hook_fn))
        return hooks

    def component_hook_fn(self, activs, hook:HookPoint, CTYPE_SCALARS:dict={}):
        layer, c_type, n_toks = int(hook.name.split(".")[1]), hook.name.split(".")[2],  activs.shape[1]
        tok_idcs, params_per_tok = self.get_token_pos(n_toks)
        scalars = CTYPE_SCALARS[layer] ## scalar_param already stored

        scalars = 1.0 + (self.beta * scalars)
        if len(activs.shape)==4: ## HEAD SCALARS: K, Q, V, Z
            n_heads = activs.shape[-2]
            activs[:,tok_idcs,:,:] = scalars.view(1,params_per_tok,n_heads,1) * activs[:,tok_idcs,:,:]
        else: ## ATTN, MLP, RESID...
            activs[:,tok_idcs,:] = scalars.view(1, params_per_tok, 1) * activs[:,tok_idcs,:]
        return activs

    def get_token_pos(self, n_toks:int):
        if self.tok_pos == "full":
            tok_idcs = torch.arange(0,n_toks).long()
            params_per_tok = n_toks
        elif self.tok_pos == "weights":
            tok_idcs = torch.arange(0,n_toks).long()    
            params_per_tok = 1
        elif self.tok_pos == "last":
            tok_idcs = torch.LongTensor([-1])  
            params_per_tok = 1
        return tok_idcs, params_per_tok

    def normalize_scalars(self, model, toks_NI, norm_type:str="max"):
        if norm_type == "max":
            self.SCALARS = {k: v / torch.max(torch.abs(v)) for k, v in self.SCALARS.items()}
        elif norm_type == "sum":
            self.SCALARS = {k: v / torch.sum(torch.abs(v)) for k, v in self.SCALARS.items()}
        self.hooks = self.register_hooks()
        self.set_hooks(model, toks_NI)

        

## DYNAMIC HOOKS____________________________________________________________________________

class DynamicScalarHooks():
    def __init__(self, model:int, init_std:float=0.01, c_types:list=["mlp_out", "attn_out"], tok_pos:str="full"):
        model.reset_hooks()
        self.device, self.n_layers, self.n_heads, self.d_model, self.d_head = model.cfg.device, model.cfg.n_layers, model.cfg.n_heads, model.cfg.d_model, model.cfg.d_head
        self.c_types, self.tok_pos, self.beta, self.init_std = c_types, tok_pos, 1.0, init_std
        self.PARAMS, self.SCALARS = self.initialize_params()
        self.hooks = self.register_hooks()
        self.set_hooks(model)

    def set_hooks(self,model):
        with torch.no_grad():
            mock_toks_NI = torch.randint(0,model.cfg.d_vocab,(1,10),requires_grad=False, device=self.device)
            orig_logits_NIT = model.run_with_hooks(mock_toks_NI, fwd_hooks=self.hooks, return_type="logits", reset_hooks_end=True)
   
    def register_hooks(self,):  
        hooks = []
        for layer in range(0,self.n_layers):
            for c_type in self.c_types:
                hook_name = transformer_lens.utils.get_act_name(c_type, layer)
                partial_hook_fn = functools.partial(self.component_hook_fn, CTYPE_PARAMS=self.PARAMS[c_type], CTYPE_SCALARS=self.SCALARS[c_type]) 
                hooks.append((hook_name, partial_hook_fn))
        return hooks

    def initialize_params(self,):
        PARAMS, SCALARS = {c_type: [] for c_type in self.c_types},  {c_type: [] for c_type in self.c_types}
        for c_type in self.c_types:
            for layer in range(0,self.n_layers):
                hook_name = transformer_lens.utils.get_act_name(c_type, layer)
                component_name = hook_name.split(".")[-1]
                if component_name in ["hook_q","hook_k","hook_v","hook_z"]: ## HEAD SCALARS: K, Q, V, Z
                    mean_1HD1 = torch.ones((1, self.n_heads, self.d_head,1),device=self.device)
                    params_1HD1 = torch.normal(mean_1HD1, self.init_std)#torch.nn.functional.normalize(torch.normal(mean_1HD1, self.init_std), p=2, dim=-1)
                    #params_1HD1 = torch.nn.Parameter(params_1HD1, requires_grad=True)
                    scalars_placeholder = None
                    PARAMS[c_type].append(params_1HD1), SCALARS[c_type].append(scalars_placeholder)
                elif component_name in ["hook_attn_out", "hook_mlp_out", "hook_resid_post", "hook_resid_pre"]: ## ATTN, MLP, RESID...
                    mean_1D1 = torch.ones((1,self.d_model,1),device=self.device)
                    params_1D1 = torch.normal(mean_1D1, self.init_std)#torch.nn.functional.normalize(torch.normal(mean_1D1, self.init_std), p=2, dim=-1)
                    #params_1D1 = torch.nn.Parameter(params_1D1, requires_grad=True)
                    scalars_placeholder = None ## scalars are just placeholders anyways
                    PARAMS[c_type].append(params_1D1), SCALARS[c_type].append(scalars_placeholder)
            PARAMS[c_type] = torch.nn.Parameter(torch.stack(PARAMS[c_type]), requires_grad=True)
        return PARAMS, SCALARS
        
    
    def component_hook_fn(self, activs, hook:HookPoint, CTYPE_PARAMS:dict={}, CTYPE_SCALARS:dict={}):
        layer, component_name, n_toks = int(hook.name.split(".")[1]), hook.name.split(".")[-1],  activs.shape[1]
        tok_idcs, params_per_tok = self.get_token_pos(n_toks)
        ctype_params = CTYPE_PARAMS[layer]

        if component_name in ["hook_q","hook_k","hook_v","hook_z"]: ## HEAD SCALARS
            activs_norm = torch.nn.functional.normalize(activs[:,tok_idcs,:,:], p=2, dim=-1)
            scalars = torch.einsum('iHDx, NIHD->NIHx', ctype_params, activs_norm)
            CTYPE_SCALARS[layer] = scalars ## store for regularization 
            scalars_one = (1.0 + (self.beta * CTYPE_SCALARS[layer]))
            activs[:,tok_idcs,:,:] = activs[:,tok_idcs,:,:] * scalars_one
        elif component_name in ["hook_attn_out", "hook_mlp_out", "hook_resid_post", "hook_resid_pre"]: ## ATTN, MLP, RESID
            activs_norm = torch.nn.functional.normalize(activs[:,tok_idcs,:], p=2, dim=-1)
            scalars = torch.einsum('iDx, NID->NIx', ctype_params, activs_norm)
            CTYPE_SCALARS[layer] = scalars ## store for regularization 
            scalars_one = (1.0 + (self.beta * CTYPE_SCALARS[layer]))
            activs[:,tok_idcs,:] = activs[:,tok_idcs,:] * scalars_one
        return activs

    def get_token_pos(self, n_toks:int):
        if self.tok_pos == "full":
            tok_idcs = torch.arange(0,n_toks).long()
            params_per_tok = n_toks
        elif self.tok_pos == "last":
            tok_idcs = torch.LongTensor([-1])  
            params_per_tok = 1
        return tok_idcs, params_per_tok

    def normalize_scalars(self, model, norm_type:str=None):
        for ctype, ctype_scalars in self.SCALARS.items():
            ctype_sum = 0
            ctype_scalars_normed = []
            for ctype_scalars_layer in ctype_scalars: ## gather sum
                ctype_sum = torch.abs(ctype_scalars_layer).sum()
            for ctype_scalars_layer in ctype_scalars: ## normalize
                ctype_scalars_layer = ctype_scalars_layer / ctype_sum
                ctype_scalars_normed.append(ctype_scalars_layer)
            self.SCALARS[ctype] = ctype_scalars_normed


        
