import torch, transformer_lens, functools, tqdm

from componentScale import trainHandler

from transformer_lens import (
    HookedTransformer,
    HookedTransformerConfig,
    FactoredMatrix,
    ActivationCache,
)

## PATCHING METHODS_________________________________________________________________


class CacheRuns:

    def __init__(self, model, toks_NI, answ_toks_N2, c_types:list=["resid_pre"], corrupt_method=("noise",0.1)):
        if isinstance(toks_NI, list):
            try:
                toks_NI = torch.stack(toks_NI)
                answ_toks_N2 = torch.stack(answ_toks_N2)
            except:
                raise ValueError(f'toks_NI all need to have same I')
        self.clean_logit_diff, self.clean_fwd_cache, self.clean_bwd_cache = self.get_cache_fwd_and_bwd(model, toks_NI, answ_toks_N2, c_types)
        print(f"clean_logit_diff: {self.clean_logit_diff}")

        if corrupt_method is not None:
            corrupt_fn = functools.partial(self.add_corrupt_hooks, corrupt_method=corrupt_method)
            self.corrupt_logit_diff, self.corrupt_fwd_cache, self.corrupt_bwd_cache = self.get_cache_fwd_and_bwd(model, toks_NI, answ_toks_N2, c_types, corrupt_fn=corrupt_fn)
            self.metric_fn = lambda logit_diff: self.metric(logit_diff)
            print(f"corrupt_logit_diff: {self.corrupt_logit_diff}")

    def metric(self, logit_diff):
        if self.corrupt_logit_diff > self.clean_logit_diff: 
            metric_val = (logit_diff-self.clean_logit_diff) / abs(self.clean_logit_diff-self.corrupt_logit_diff)
        else:
            metric_val = (self.clean_logit_diff-logit_diff) / abs(self.clean_logit_diff-self.corrupt_logit_diff)
        return metric_val

    
    def add_corrupt_hooks(self, model, components:list=[("embed", None)], corrupt_method:tuple=("noise",0.1)):
        def forward_corrupt_hook(act, hook, corrupt_method):
            assert len(corrupt_method), f"corrupt_method {corrupt_method} needs to be tuple"
            act = act.detach()
            if corrupt_method[0]=="mean":
                act = torch.mean(act).expand_as(act)
            if corrupt_method[0]=="zero":
                act = torch.zeros(act.shape, device=act.device)
            if corrupt_method[0]=="noise":
                act = act + torch.normal(torch.zeros(act.shape, device=act.device), corrupt_method[1])
            return act
    
        for (component_name, layer) in components:
            hook_name = transformer_lens.utils.get_act_name(component_name, layer)
            forward_corrupt_fn = functools.partial(forward_corrupt_hook, corrupt_method=corrupt_method)
            model.add_hook(hook_name, forward_corrupt_fn, "fwd")
            print(f"added corruption at {hook_name}")
        
    def get_cache_fwd_and_bwd(self, model, toks_NI, answ_toks_N2, c_types:list=["mlp_out", "attn_out"], corrupt_fn=None):
    
        model.reset_hooks()
        torch.set_grad_enabled(True)
        def forward_cache_hook(act, hook, CTYPE_DICT):
            CTYPE_DICT.append(act.detach())
        def backward_cache_hook(act, hook, CTYPE_DICT):
            CTYPE_DICT.append(act.detach())
    
        if callable(corrupt_fn):
            corrupt_fn(model)
            
        fwd_cache, bwd_cache = {}, {}
        for c_type in c_types:
            fwd_cache[c_type], bwd_cache[c_type] = [], []
            for layer in range(0, model.cfg.n_layers):
                hook_name = transformer_lens.utils.get_act_name(c_type, layer)
                model.add_hook(hook_name, functools.partial(forward_cache_hook,CTYPE_DICT=fwd_cache[c_type]), "fwd")
                model.add_hook(hook_name, functools.partial(forward_cache_hook,CTYPE_DICT=bwd_cache[c_type]), "bwd")
        logits_NIT = model(toks_NI)
        logit_diff, (logit_tok0, logit_tok1) = trainHandler.get_logit_diff(logits_NIT, answ_toks_N2=answ_toks_N2)
        #logit_diff = logit_diff / (logit_tok0 + logit_tok1)
        logit_diff = logit_diff.mean()
    
        logit_diff.backward()
        model.reset_hooks()
    
        for c_type in c_types: ## change to shape --> N,I,L,D
            fwd_cache[c_type] = torch.stack(fwd_cache[c_type])
            fwd_cache[c_type] = torch.swapaxes(fwd_cache[c_type], 0, 1)
            
            bwd_cache[c_type] = torch.stack(bwd_cache[c_type])
            bwd_cache[c_type] = torch.swapaxes(bwd_cache[c_type], 0, 1)
        return logit_diff.item(), fwd_cache, bwd_cache
    


## ACTIVATION PATCHING________________________________________________________________________________

class ActivPatcher:

    def __init__(self, model, toks_NI, answ_toks_N2, corrupt_fwd_cache, metric_fn=None):
        if isinstance(toks_NI, list):
            try:
                toks_NI = torch.stack(toks_NI)
                answ_toks_N2 = torch.stack(answ_toks_N2)
            except:
                raise ValueError(f'toks_NI all need to have same I')
        self.metric_fn = metric_fn
        self.SCALARS = self.run_activation_patching(model, toks_NI, answ_toks_N2, corrupt_fwd_cache)

    def patching_hook(self, act, hook, patch_activs, tok_idx:int):
        layer = int(hook.name.split(".")[1])
        act[:,tok_idx,:] = patch_activs[:,layer,tok_idx,:]
        return act
    
    def run_activation_patching(self, model, toks_NI, answ_toks_N2, corrupt_fwd_cache):
        SCALARS, n_layers, n_toks = {}, model.cfg.n_layers, toks_NI[0].shape[0]
        torch.set_grad_enabled(False)
        for c_type, corrupt_fwd in corrupt_fwd_cache.items():
            SCALARS[c_type] = torch.zeros(n_layers, n_toks)
            for layer in tqdm.tqdm(range(0, n_layers)):
                for tok_idx in range(0, n_toks):
                    hook_name = transformer_lens.utils.get_act_name(c_type, layer)
                    partial_hook_fn = functools.partial(self.patching_hook,patch_activs=corrupt_fwd, tok_idx=tok_idx)
                    logits_NIT = model.run_with_hooks(toks_NI, fwd_hooks=[(hook_name, partial_hook_fn)], reset_hooks_end=True)
                    logit_diff, (logit_tok0, logit_tok1) = trainHandler.get_logit_diff(logits_NIT, answ_toks_N2=answ_toks_N2)
                    if self.metric_fn is not None:
                        logit_diff = self.metric_fn(logit_diff)
                    SCALARS[c_type][layer, tok_idx] = logit_diff.mean(0)
            ## normalize to [-1, 1] range
            abs_max_val = torch.max(torch.abs(SCALARS[c_type])) 
            SCALARS[c_type] = SCALARS[c_type] / abs_max_val
        return SCALARS
        

## ATTRIBUTION PATCHING________________________________________________________________________________

class AttrPatcher:

    def __init__(self, cr):
        clean_fwd_cache, clean_bwd_cache, corrupt_fwd_cache = cr.clean_fwd_cache, cr.clean_bwd_cache, cr.corrupt_fwd_cache
        self.clean_logit_diff, self.corrupt_logit_diff = cr.clean_logit_diff, cr.corrupt_logit_diff
        self.SCALARS = self.get_attribution_scores(clean_fwd_cache, clean_bwd_cache, corrupt_fwd_cache)

    def get_attribution_scores(self, clean_fwd_cache, clean_bwd_cache, corrupt_fwd_cache, mean_N:bool=True):
        SCALARS = {}
        for c_type, clean_fwd in clean_fwd_cache.items():
            clean_bwd, corrupt_fwd = clean_bwd_cache[c_type], corrupt_fwd_cache[c_type]
            #clean_fwd[:,-1] = torch.nn.functional.normalize(clean_fwd[:,-1], p=2, dim=-1)
            #corrupt_fwd[:,-1] = torch.nn.functional.normalize(corrupt_fwd[:,-1], p=2, dim=-1)
            #clean_bwd[:,-1] = torch.nn.functional.normalize(clean_bwd[:,-1], p=2, dim=-1)
            if self.clean_logit_diff < self.corrupt_logit_diff:
                diff_fwd = corrupt_fwd-clean_fwd
            else:
                diff_fwd = clean_fwd-corrupt_fwd

            ## normalization
            #fwd_layer_norm = (1 / torch.norm(diff_fwd, p='fro', dim=-1))
            #fwd_layer_norm = (1 / torch.linalg.vector_norm(diff_fwd, ord=2, dim=-1)) 
            if len(corrupt_fwd.shape) > 4: ## head 
                #diff_fwd = torch.einsum("NLIHD, NLHI -> NLIHD", diff_fwd, fwd_layer_norm)
                c_type_attr = torch.einsum("NLIHD, NLIHD -> NLIH", diff_fwd, clean_bwd)
            else:
                #diff_fwd = torch.einsum("NLID, NLI -> NLID", diff_fwd, fwd_layer_norm)
                c_type_attr = torch.einsum("NLID, NLID -> NLI", diff_fwd, clean_bwd)

            ## normalization
            #c_type_attr = torch.nn.functional.normalize(c_type_attr, p=1, dim=-1)
            fwd_layer_norm = torch.linalg.vector_norm(diff_fwd, ord=float("inf"), dim=-1) #float("inf")
            c_type_attr = c_type_attr * (1 / fwd_layer_norm)
            SCALARS[c_type] = c_type_attr
            if mean_N: ## mean over prompts
                SCALARS[c_type] = SCALARS[c_type].mean(0) 
        return SCALARS


## DIRECT LOGIT ATTRIBUTION_____________________________________________


class DLA:
    def __init__(self, model, toks_NI, answ_toks_N2, c_types:list=["resid_pre"], tok_pos:str="full"):

        answ_dirs_ND = self.get_logit_diff_direction(model, answ_toks_N2)
        self.SCALARS = self.get_model_dla(model, toks_NI, answ_dirs_ND, c_types, tok_pos)

    def get_logit_diff_direction(self, model, answ_toks_N2:torch.Tensor):
        #unembed_vecs = model.tokens_to_residual_directions(scale_ids) # model.unembed.W_U
        logit_diff_dir_ND = torch.empty([len(answ_toks_N2), model.cfg.d_model])
        for n, answ_toks_2 in enumerate(answ_toks_N2):
            unembed_vecs = torch.index_select(model.unembed.W_U, dim=-1, index=answ_toks_2.view(-1))
            unembed_vecs = torch.swapaxes(unembed_vecs, 0, 1)
            unembed_vecs = unembed_vecs.view(1, 2, model.cfg.d_model)
            logit_diff_dir_ND[n,:] = unembed_vecs[...,0,:] - unembed_vecs[...,1,:]
        logit_diff_dir_nD = logit_diff_dir_ND.mean(0).unsqueeze(0) ## mean over all answer token directions
        return logit_diff_dir_nD
    
    def get_model_dla(self, model, toks_NI, answ_dirs_ND, c_types:list=["resid_pre"], tok_pos:str="full", mean_N:bool=True):
        if isinstance(toks_NI, list):
            try:
                toks_NI = torch.stack(toks_NI)
            except:
                raise ValueError(f'toks_NI all need to have same I')
        _, cache = model.run_with_cache(toks_NI)
        SCALARS = {c_type: [] for c_type in c_types}
        for c_type in c_types:
            for layer in range(0,model.cfg.n_layers):
                component = transformer_lens.utils.get_act_name(c_type, layer)
                h_NID = model.ln_final(cache[component])[:, :, :]
                #print(component, h_NID.shape, answ_dirs_ND.shape)
                if len(h_NID.shape) > 3: ## head 
                    #model.W_O ## n_layers, n_heads, d_head, d_model
                    head_attr_NIHD = torch.einsum('...NIHM,...HMD->NIHD', h_NID, model.W_O[layer])
                    attr_NI = torch.einsum('...NIHD,...ND->NIH', head_attr_NIHD, answ_dirs_ND)
                else: ## resid_pre, mlp_out, attn_out ....
                    attr_NI = torch.einsum('...NID,...ND->NI', h_NID, answ_dirs_ND)
                SCALARS[c_type].append(attr_NI)
            SCALARS[c_type] = torch.swapaxes(torch.stack(SCALARS[c_type]),0,1) ## swap N and layers
            if mean_N: ## mean over prompts
                SCALARS[c_type] = SCALARS[c_type].mean(0)  
            if tok_pos=="last":
                SCALARS[c_type][:,:-1] = 0
        return SCALARS

