import torch, gc, os, transformer_lens, itertools, tqdm, logging, dotenv, itertools, pathlib
dotenv.load_dotenv()
logger = logging.getLogger('logger')
os.environ["TOKENIZERS_PARALLELISM"] = "false"

def set_logger(level:str="INFO"):
    """
    set up logger
    """
    logger = logging.getLogger('logger')
    logger.setLevel({"DEBUG": logging.DEBUG, "INFO": logging.INFO}[level])
    logger.addHandler(logging.StreamHandler())


def get_device(device:str=None):
    """
    return devices
    """
    mps_available = [["","mps"][int(hasattr(torch.backends, "mps") and torch.backends.mps.is_available())]]
    cuda_available = [["","cuda"][int(torch.cuda.is_available())]]
    devices = [device for device in mps_available + cuda_available + ["cpu"] if device != ""]
    if isinstance(device, str):
        if device in devices: ## return desired device
            return device  
    return devices[0]


def load_model(model_type:str="gpt2-small", DEVICE:str="cpu", freeze_params:bool=True):
    """
    load pretrained model (and optimizer)
    """
    model = transformer_lens.HookedTransformer.from_pretrained(model_type, device=get_device(DEVICE))
    try:
        model.set_use_attn_result(True)
        model.set_use_attn_in(True)
        model.set_use_hook_mlp_in(True)
    except:
        pass

    set_no_grad(model, freeze_params) ## set no grad
    return model


def gpu_check(print_stats:bool=True):
    """
    GPU checking and cleaning
    """
    gc.collect()
    torch.cuda.empty_cache()
    if print_stats:
        conv_byte2gb = 1000000000
        t = round(torch.cuda.get_device_properties(0).total_memory / conv_byte2gb, 4)
        r = round(torch.cuda.memory_reserved(0) / conv_byte2gb, 4)
        a = round(torch.cuda.memory_allocated(0) / conv_byte2gb, 4)
        print(f"Gb total: {t}, reserved: {r}, allocated: {a}")


def set_no_grad(model, freeze_params:bool=True):
    """
    explicitely set or remove parameters that do not require gradient
    """
    #"embed", "pos_embed", "unembed", "b_in", "b_out", "b_K", "b_Q", "b_V", "b_O"
    #torch.set_grad_enabled(grad)
    if freeze_params:
        for name, param in model.named_parameters():
            #logger.info(f"setting no_grad on {no_grad}")
            param.requires_grad = False
        logger.info(f"no grad on model params")


def NegLogLik(logits_T:torch.FloatTensor):
    """
    standard negative log likelihood loss for language modeling
    """
    log_probs_T = -(logits_T.log_softmax(dim=-1)) ## negative log probs 
    return log_probs_T


def gather_token_scores(scores_NIT:torch.FloatTensor, labels_NI:torch.LongTensor=None):
    """
    select scores corresponding to sequence of tokens (NIT)
    """     
    if len(labels_NI.shape) == 2: ## single sequence of labels
        scores = scores_NIT.gather(dim=-1, index=labels_NI[:, 1:, None])[:, :, 0] 
    elif len(labels_NI.shape) == 3: ## mutiple sequences of labels
          scores = scores_NIT.gather(dim=-1, index=labels_NI[:, 1:])
    return scores

def get_topK(scores_NIT:torch.FloatTensor, topK:int=None, minK:bool=False):
    """
    return the topK scores and token indeces
    """
    if topK is None:
        topK = scores_NIT.shape[-1]
    scores_NIk, indc_NIk = torch.topk(scores_NIT, topK, dim=-1, largest= not minK)
    return scores_NIk, indc_NIk
