import torch, logging, tqdm
from torch.utils.data import DataLoader
logger = logging.getLogger('logger')

## METRICS______________________________________________

KLDivLoss = torch.nn.KLDivLoss(reduction="none", log_target=True)
RegLoss =  torch.nn.L1Loss() #torch.nn.SmoothL1Loss()# torch.nn.HuberLoss(reduction='mean', delta=1.0) #RegLoss = torch.nn.MSELoss()

class JSD(torch.nn.Module):
    def __init__(self):
        super(JSD, self).__init__()
        self.kl = torch.nn.KLDivLoss(reduction='batchmean', log_target=True)

    def forward(self, p: torch.tensor, q: torch.tensor):
        p, q = p.view(-1, p.size(-1)), q.view(-1, q.size(-1))
        m = (0.5 * (p + q)).log()
        return 0.5 * (self.kl(p.log(), m) + self.kl(q.log(), m))

jsd = JSD()

def regularize_scalars(params:list, reg_to_val:float=0):
    reg_loss = 0
    for param in params:
        if reg_to_val==1:
            keep_vals = torch.ones(param.shape,requires_grad=False,device=param.device)
        elif reg_to_val==0:
            keep_vals = torch.zeros(param.shape,requires_grad=False,device=param.device)
        param_reg = RegLoss(param, keep_vals)
        reg_loss += param_reg.mean()
    return reg_loss


def get_logit_diff(logits, answ_toks_N2:torch.LongTensor, orig_logits_NIT=None, norm_types:list=[""]):
    answ_toks_N2 = answ_toks_N2.to(logits.device)
    logits_answ0 = logits[...,-1,:].gather(-1,answ_toks_N2[...,0, None])
    logits_answ1 = logits[...,-1,:].gather(-1,answ_toks_N2[...,1, None])

    if orig_logits_NIT is not None:
        orig_logits_answ0 = orig_logits_NIT[...,-1,:].gather(-1,answ_toks_N2[...,0, None])
        orig_logits_answ1 = orig_logits_NIT[...,-1,:].gather(-1,answ_toks_N2[...,1, None])

        if orig_logits_answ0 > orig_logits_answ1:
            logit_diff = (logits_answ0 - logits_answ1)
        else:
            logit_diff = (logits_answ1 - logits_answ0)
    else: 
        logit_diff = (logits_answ0 - logits_answ1)
        
    #assert logit_diff < 0, f"flip token order {answ_toks_N2} logits {logit_diff}"
    if "abs" in norm_types: ## get logit difference in absolute values
        logit_diff = torch.abs(logit_diff)
    if "power_two" in norm_types:
        logit_diff = logit_diff**2
    if "divide_sum" in norm_types:
        logit_diff = logit_diff / (torch.abs(logits_answ0) + torch.abs(logits_answ1))
    if "divide_max" in norm_types:
        logit_diff = logit_diff / torch.max(torch.abs(logits_answ0), torch.abs(logits_answ1))
    return logit_diff, (logits_answ0, logits_answ1)


def get_kl_div(logits_NIT, orig_logits_NIT:torch.FloatTensor, answ_toks_N2:torch.LongTensor):
    ## exclude contrastive pairs from KL divergence loss 
    
    if len(answ_toks_N2.shape) == 1:
        answ_toks_N2 = answ_toks_N2.unsqueeze(0).repeat_interleave(logits_NIT.shape[0], dim=0)

    mask_NT = torch.zeros(logits_NIT[:,-1,:].shape, device=logits_NIT.device)
    mask_NT = ~mask_NT.scatter_(1, answ_toks_N2, 1).bool()
    
    logits_masked = logits_NIT[:,-1,:][mask_NT]
    orig_logits_masked = orig_logits_NIT[:,-1,:][mask_NT]

    input = torch.nn.functional.softmax(logits_masked, dim=-1)
    target = torch.nn.functional.softmax(orig_logits_masked, dim=-1) 
    #input = torch.nn.functional.log_softmax(logits_masked, dim=-1)
    #target = torch.nn.functional.log_softmax(orig_logits_masked, dim=-1)    
    #kl_div = KLDivLoss(input, target)
    kl_div = jsd(input, target)
    kl_div = kl_div.mean()
    #assert kl_div >= 0, f"kl divergence must be >= 0, but is {kl_div}"
    return kl_div


## TRAINING LOOP______________________________________________

def get_orig_logits(model, toks_NI:list, answ_toks_N2:list, norm_types):
    reduce_single_logit = None
    with torch.no_grad():
        orig_logits_NIT, orig_logit_diffs = [], []
        for n, (toks_I, answ_toks_2) in enumerate(zip(toks_NI, answ_toks_N2)):
            toks_nI, answ_toks_n2 = toks_I.unsqueeze(0), answ_toks_2.unsqueeze(0)
            orig_logits_nIT, _ = model.run_with_cache(toks_nI.detach())
            orig_logit_diff, (orig_logit_answ0, orig_logit_answ1) = get_logit_diff(orig_logits_nIT, answ_toks_n2, None, norm_types)
            if isinstance(reduce_single_logit, int): ## reducing a single logit, either 0 or 1
                orig_logit_diff = (orig_logit_answ0, orig_logit_answ1)[reduce_single_logit]
            orig_logits_NIT.append(orig_logits_nIT), orig_logit_diffs.append(orig_logit_diff)
            print(f"{n+1}/{len(toks_NI)}: orig_logit_diff: {round(orig_logit_diff.mean().item(),3)}, answ: {model.to_str_tokens(answ_toks_n2.view(-1))}")    
    return orig_logits_NIT, orig_logit_diffs
    

def trainHooks(sh, model, optim, hyperp:dict, toks_NI, answ_toks_N2, batch_size:int=1, force_logits_apart:bool=True, norm_types=[""]):
    reduce_single_logit, reg_to_val= None, 0
    orig_logits_NIT, orig_logit_diffs = get_orig_logits(model, toks_NI, answ_toks_N2, norm_types)
    print(f"\nstarting the training, force_logits_apart:{force_logits_apart}, norm_types:{norm_types}, reg_to_val:{reg_to_val}")
    torch.set_grad_enabled(True)

    collate_fn = lambda x: x
    dl = DataLoader(list(zip(toks_NI, answ_toks_N2, orig_logits_NIT, orig_logit_diffs)), batch_size=batch_size, shuffle=True, collate_fn=collate_fn)

    for epoch_i in range(0,hyperp["epochs"]):
        for batch_i, batch in enumerate(dl):
            logit_diffs, kl_losses, orig_logit_diffs = 0,0,0
            for (toks_I, answ_toks_2, orig_logits_NIT, orig_logit_diff) in batch:
                toks_nI, answ_toks_n2 = toks_I.unsqueeze(0), answ_toks_2.unsqueeze(0)
                interv_logits_NIT = model.run_with_hooks(toks_nI.detach(), fwd_hooks=sh.hooks, return_type="logits", reset_hooks_end=True)        
                logit_diff, (logits_answ0, logits_answ1) = get_logit_diff(interv_logits_NIT, answ_toks_n2, orig_logits_NIT, norm_types=norm_types)
                
                if force_logits_apart: ## force answer token logits apart
                    logit_diff = -logit_diff
                if isinstance(reduce_single_logit, int): ## reducing a single logit, either 0 or 1
                    logit_diff = (logits_answ0, logits_answ1)[reduce_single_logit]
                    kl_loss = get_kl_div(interv_logits_NIT, orig_logits_NIT.detach(), answ_toks_n2[:,reduce_single_logit])
                else:
                    kl_loss = get_kl_div(interv_logits_NIT, orig_logits_NIT.detach(), answ_toks_n2)
                
                logit_diffs += logit_diff
                orig_logit_diffs += orig_logit_diff
                kl_losses += kl_loss
                
            
            ## COMPUTE LOSS__________________________________________
            logit_diff = logit_diffs / len(batch)
            orig_logit_diff = orig_logit_diffs / len(batch)
            kl_loss = kl_losses / len(batch)
            
            if hasattr(sh, 'PARAMS') and hasattr(sh, 'SCALARS'):
                ctype_scalars = [torch.stack(ctype_scalars) for ctype, ctype_scalars in sh.SCALARS.items()]
                reg_loss = regularize_scalars(ctype_scalars, reg_to_val=reg_to_val) ## regularize scalars, not params
            elif hasattr(sh, 'SCALARS'):
                reg_loss = regularize_scalars(optim.param_groups[0]["params"], reg_to_val=reg_to_val) 

            loss = logit_diff + (hyperp["kl_lambda"] * kl_loss) + (hyperp["reg_lambda"] * reg_loss)
            print(f"batch {batch_i+1}/{len(dl)}, epoch {epoch_i+1}/{hyperp['epochs']} loss {round(loss.item(),3)}, logit_diff: {round(logit_diff.mean().item(),3)}<-{round(orig_logit_diff.mean().item(),3)}, kl_loss {round(kl_loss.mean().item(),3), round((hyperp['kl_lambda'] * kl_loss).item(),3)}, reg_loss {round(reg_loss.item(),3), round((hyperp['reg_lambda'] * reg_loss).item(),3)}")
            
            loss.backward(retain_graph=False)
            optim.step()
            optim.zero_grad()
            model.zero_grad() 
            
